"""# Auto README

Tool for automatically generating READMEs based on the kind of repository that
it is invoked on.
"""

import dataclasses
import enum
import logging
import subprocess
from collections.abc import Iterator
from pathlib import Path

import bs4

from ._version import __version__  # noqa: F401
from .markdown import concat_markdown, html_to_markdown

logger = logging.getLogger(__name__)

JAVADOC_URL_PREFIX = 'https://secata.gitlab.io'

DOCKER_USAGE = """
## Usage

To build a local docker image, run:

```sh
mvn package com.google.cloud.tools:jib-maven-plugin:dockerBuild -pl {ARTIFACT_ID} -am
```

Then to run the docker image:

```sh
sudo docker run -v CONFIG_PATH_HERE":/conf/server.json" {ARTIFACT_ID}:latest
```

See <conf> directory for example configurations.
"""


SRC_ROOT = Path('/home/jmaa/src/secata')  # TODO


class ProjectType(enum.Enum):
    JAVA_CONTRACT = 'JAVA_CONTRACT'
    JAVA_DOCKER = 'JAVA_DOCKER'
    JAVA_EXECUTABLE = 'JAVA_EXECUTABLE'
    JAVA_LIBRARY = 'JAVA_LIBRARY'


@dataclasses.dataclass(frozen=True)
class ProjectSetup:
    project_type: ProjectType
    main_class_path: str | None


def add_further_links(
    markdown_out: list[str],
    further_links: list[tuple[str, str]],
) -> None:
    markdown_out.append('## Further links\n')
    for text, url, extra_text in further_links:
        markdown_out.append(f'- [{text}]({url})')
        if extra_text is not None:
            markdown_out.append(': ')
            markdown_out.append(extra_text)
        markdown_out.append('\n')


def parse_pom(project_path: Path):
    with open(project_path / 'pom.xml') as f:
        return bs4.BeautifulSoup(f, 'lxml-xml')


DEPENDENCY_FORMAT = """
<dependency>
    <groupId>{GROUP_ID}</groupId>
    <artifactId>{ARTIFACT_ID}</artifactId>
    <version>{VERSION}</version>
</dependency>
""".strip()


def get_maven_group_artifact_id(pom) -> tuple[str, str]:
    group_id = pom.find('groupId').get_text().strip()
    artifact_id = pom.find('artifactId').get_text().strip()
    return (group_id, artifact_id)


def usage_maven_import_library(markdown_out: list[str], project_path: Path):
    (group_id, artifact_id) = get_maven_group_artifact_id(parse_pom(project_path))

    markdown_out.append('## Usage, Maven\n')
    markdown_out.append(
        'In your Maven `pom.xml`, add following to the `<dependencies>` section:\n',
    )
    markdown_out.append('```xml\n')
    markdown_out.append(
        DEPENDENCY_FORMAT.format(
            GROUP_ID=group_id,
            ARTIFACT_ID=artifact_id,
            VERSION='VERSION',
        ),
    )
    markdown_out.append('\n```\n\n')
    markdown_out.append('A suitable `VERSION` can be found in the package registry.\n')


def readme_for_repo_without_main(markdown_out: list[str]):
    markdown_out.append('Java Library.\n\n')


def path_to_classpath(path: Path) -> str:
    return str(path).removesuffix('.java').replace('.', '/')


def classpath_to_path(classpath: str) -> str:
    return str(classpath).replace('.', '/')


def readme_from_class(project_path: Path, project_setup: ProjectSetup) -> str:
    main_class_doc_soup = None
    if project_setup.main_class_path is not None:
        docs_rel_path = classpath_to_path(project_setup.main_class_path)

        docs_file = (
            project_path / 'target' / 'reports' / 'apidocs' / docs_rel_path
        ).with_suffix('.html')
        with open(docs_file) as f:
            main_class_doc_soup = bs4.BeautifulSoup(f, 'lxml')

    title = project_path.name.replace('-', ' ').capitalize()

    gitlab_project_path = project_path.relative_to(SRC_ROOT)

    javadoc_url = JAVADOC_URL_PREFIX + '/' + str(gitlab_project_path)
    further_links = [('Javadoc Documentation for `main` branch', javadoc_url, None)]

    markdown_out = ['# ', title, '\n\n']
    if main_class_doc_soup is not None:
        html_to_markdown(
            markdown_out,
            main_class_doc_soup.select_one('#class-description .block'),
        )
        see_elems = main_class_doc_soup.select(
            'ul.see-list-long li',
        ) or main_class_doc_soup.select('ul.tag-list-long li')
        logger.info('Found %d @see elements', len(see_elems))
        for see_elem in see_elems:
            a = see_elem.a.extract()
            further_links.append((a.get_text(), a['href'], see_elem.get_text()))
            del see_elem, a
        del see_elems
    else:
        readme_for_repo_without_main(markdown_out)
        usage_maven_import_library(markdown_out, project_path)

    if project_setup.project_type == ProjectType.JAVA_DOCKER:
        (group_id, artifact_id) = get_maven_group_artifact_id(parse_pom(project_path))
        markdown_out += [DOCKER_USAGE.format(ARTIFACT_ID=artifact_id)]

    add_further_links(markdown_out, further_links)
    return concat_markdown(markdown_out)


def files_containing(text: str, path: Path):
    result = subprocess.run(['grep', text, '-r', path, '-l'], capture_output=True)
    results = result.stdout.strip().split(b'\n')
    return [Path(r.strip().decode('utf8')) for r in results if r.strip() != b'']


def project_setup_for_contract_maven_java(project_path: Path) -> ProjectSetup | None:
    # main resource
    main_resource_path = project_path / 'src' / 'main' / 'resources' / 'main'
    if main_resource_path.exists():
        with open(main_resource_path) as f:
            main_class_path = f.read().strip()

        return ProjectSetup(ProjectType.JAVA_CONTRACT, main_class_path)

    # AutoSysContract
    main_java_path = project_path / 'src' / 'main' / 'java'
    if system_contract_files := files_containing('@AutoSysContract', main_java_path):
        assert len(system_contract_files) == 1
        main_class_path = path_to_classpath(
            system_contract_files[0].relative_to(main_java_path),
        )
        return ProjectSetup(ProjectType.JAVA_CONTRACT, main_class_path)

    # Main method
    if main_method_files := files_containing('public static void main', main_java_path):
        assert len(main_method_files) == 1
        main_class_path = path_to_classpath(
            main_method_files[0].relative_to(main_java_path),
        )
        return ProjectSetup(ProjectType.JAVA_EXECUTABLE, main_class_path)

    return None


def project_setup_for_docker_maven_java(project_path: Path) -> ProjectSetup | None:
    pom = parse_pom(project_path)
    class_soup = pom.find('configurator.class')
    if class_soup is None:
        return None

    main_class_path = class_soup.get_text().strip()

    return ProjectSetup(ProjectType.JAVA_DOCKER, main_class_path)


def multi_maven_modules(project_path: Path) -> Iterator[str]:
    pom = parse_pom(project_path)

    for module_soup in pom.select('modules module'):
        yield module_soup.get_text()


def auto_readme_for_partisia_maven_java_multi_mavens(project_path: Path) -> None:
    title = project_path.name.replace('-', ' ').capitalize()

    markdown_out = ['# ', title, '\n\n']
    markdown_out.append('## Repository layout\n')
    markdown_out.append('Submodules:\n')

    for submodule in multi_maven_modules(project_path):
        markdown_out.append(f'- [`{submodule}`](./{submodule})\n')
    return concat_markdown(markdown_out)


def determine_project_setup(project_path: Path) -> ProjectSetup:
    # Assume contract
    if project_setup := project_setup_for_contract_maven_java(project_path):
        return project_setup

    # Assume docker
    if project_setup := project_setup_for_docker_maven_java(project_path):
        return project_setup

    # Assume general maven java
    return ProjectSetup(ProjectType.JAVA_LIBRARY, None)


def auto_readme_for_project(project_path: Path, generate_multi_maven: bool) -> None:
    new_readme = None

    # Check whether multi-module
    submodules = list(multi_maven_modules(project_path))
    if submodules:
        for submodule in submodules:
            auto_readme_for_project(project_path / submodule)
            del submodule
        if generate_multi_maven:
            new_readme = auto_readme_for_partisia_maven_java_multi_mavens(project_path)
        return

    if new_readme is None:
        project_setup = determine_project_setup(project_path)
        logger.info('Project setup: %s', project_setup)
        new_readme = readme_from_class(
            project_path,
            project_setup,
        )

    assert new_readme is not None, project_path

    logger.info('Writing README')
    readme_path = project_path / 'README.md'
    with open(readme_path, 'w') as f:
        f.write(new_readme)
